import numpy as np
from IPython.display import HTML, display
import pandas as pd
import plotly.express as px

class nCrResult(list):
    """
    Class to store the result of a nCrEngine calculation.
    """

    def __init__(self, *args, rf_rate: float):
        super().__init__(*args)
        self.horizon_rf_rate = rf_rate


    def _beautify_portfolio(self, portfolio_dict):
        portfolio = portfolio_dict['portfolio'].split(' - ')
        weights = portfolio_dict['weights']
        expected_returns = portfolio_dict['expected_returns']
        cov_matrix = portfolio_dict['cov_matrix']
        betas = portfolio_dict['betas']

        # Create a DataFrame for the portfolio and weights
        portfolio_df = pd.DataFrame({
            "Weight": weights,
            "Expected Return": expected_returns,
            "Beta": betas
        }, index=portfolio)

        # Generate HTML for the table
        portfolio_html = portfolio_df.to_html(float_format="%.4f")
        cov = pd.DataFrame(cov_matrix, index=portfolio, columns=portfolio).to_html(float_format="%.4f")

        # Create plotly scatter of all portfolios
        all_portfolios = {
            'Portfolio': [x['portfolio'] for x in self],
            'Return': [x['return'] for x in self],
            'Variance': [x['portfolio_variance'] for x in self]
        }

        df = pd.DataFrame(all_portfolios)
        df.set_index('Portfolio', inplace=True)

        df['Sharpe Ratio'] = (df['Return'] - self.horizon_rf_rate) / np.sqrt(df['Variance'])

        max_return = df.loc[df['Return'].idxmax(), ['Return', 'Variance', 'Sharpe Ratio']]
        min_volatility = df.loc[df['Variance'].idxmin(), ['Return', 'Variance', 'Sharpe Ratio']]
        best_portfolio = df.loc[(df['Return'] / df['Variance']).idxmax(), ['Return', 'Variance', 'Sharpe Ratio']]

        fig = px.scatter(df,
                         x='Variance',
                         y='Return',
                         color='Sharpe Ratio',
                         color_continuous_scale='Viridis',
                         title='Optimized Portfolios')

        # Add traces for each key portfolio with proper colors and symbols
        fig.add_trace(
            px.scatter(
                x=[max_return['Variance']], y=[max_return['Return']],
                size=[10], symbol=[f'Max Return: {max_return.name}'], labels={'x': 'Variance', 'y': 'Return'},
            ).data[0]
        )
        fig.data[-1].marker.color = 'red'
        fig.data[-1].marker.size = 10

        fig.add_trace(
            px.scatter(
                x=[min_volatility['Variance']], y=[min_volatility['Return']],
                size=[10], symbol=[f'Min Volatility: {min_volatility.name}'], labels={'x': 'Variance', 'y': 'Return'},
            ).data[0]
        )
        fig.data[-1].marker.color = 'green'
        fig.data[-1].marker.size = 10

        fig.add_trace(
            px.scatter(
                x=[best_portfolio['Variance']], y=[best_portfolio['Return']],
                size=[10], symbol=[f'Best Portfolio: {best_portfolio.name}'], labels={'x': 'Variance', 'y': 'Return'},
            ).data[0]
        )
        fig.data[-1].marker.color = 'violet'
        fig.data[-1].marker.size = 10

        # Update layout to improve legend placement and color scale
        fig.update_layout(
            xaxis_title='Variance',
            yaxis_title='Return',
            title='Optimized Portfolios',
            coloraxis_colorbar=dict(
                title='Sharpe Ratio',
                tickvals=[min(df['Sharpe Ratio']), max(df['Sharpe Ratio'])],
                ticktext=[f'{min(df["Sharpe Ratio"]):.2f}', f'{max(df["Sharpe Ratio"]):.2f}'],
            ),
            legend=dict(
                orientation="h",
                yanchor="bottom",
                y=1.1,
                xanchor="right",
                x=0.5
            )
        )

        # Create additional HTML content for metrics
        summary_html = f"""
            <h2>Portfolio Analysis</h2>
            <h3>Summary Metrics</h3>
            <ul>
                <li><strong>Expected Portfolio Return:</strong> {portfolio_dict['return'] * 100:.4f}%</li>
                <li><strong>Portfolio Variance:</strong> {portfolio_dict['portfolio_variance'] * 100:.4f}%</li>
            </ul>
            <h3>Portfolio Composition</h3>
            {portfolio_html}

            <h3>Covariance Matrix</h3>
            {cov}
            
            <h3>Optimized Combination Space</h3>
            """

        display(HTML(summary_html))
        fig.show()

    def _best_portfolio(self) -> HTML | dict:
        return max(
            self,
            key=lambda x: x['return'] / x['portfolio_variance']
        )

    def _max_return(self, display: bool = False) -> dict:
        if display:
            return self.beautify_portfolio(self.max_return())
        return max(
            self,
            key=lambda x: x['return']
        )

    def _min_volatility(self) -> dict:
        return min(
            self,
            key=lambda x: x['portfolio_variance']
        )

    def max_return(self, display: bool = False) -> dict | HTML:
        """
        Get the maximum return from the result.
        """
        if display:
            return self._beautify_portfolio(self._max_return())
        return self._max_return()

    def min_volatility(self, display: bool = False) -> dict | HTML:
        """
        Get the minimum volatility from the result.
        """
        if display:
            return self._beautify_portfolio(self._min_volatility())
        return self._min_volatility()

    def best_portfolio(self, display: bool = False) -> dict | HTML:
        """
        Get the best portfolio from the result.
        """
        if display:
            return self._beautify_portfolio(self._best_portfolio())
        return self._best_portfolio()


import os.path

from .Portfolio import Portfolio
from .ReturnPred import ReturnPred
from .nCrEngine import nCrEngine
from .Sentiment import Sentiment
from .Cache import PortfolioCache
from .nCrResult import nCrResult
from .CustomTypes import IndexSymbol, Days

import datetime as dt

import pandas as pd
import yfinance as yf

import numpy as np
from scipy.optimize import minimize


from typing import Optional, Any
from os import PathLike

import warnings

from tqdm.auto import tqdm


class nCrOptimize(nCrEngine):
    """Find the optimal portfolio for a target return in a combination pool created from index components."""
    def __init__(self,
                 market: IndexSymbol,
                 n: int,
                 target_return: float = 0.1,
                 horizon: Days = 21,
                 lookback: Days = 252,
                 max_pool_size: Optional[int] = None,
                 api_key_path: Optional[PathLike | str] = None,
                 api_var_name: Optional[str] = None) -> None:

        super().__init__(market, n, horizon, lookback, max_pool_size, target_return)

        self.api_key_path = api_key_path
        self.key_var = api_var_name

        if isinstance(api_key_path, PathLike) and not os.path.exists(api_key_path):
            raise FileNotFoundError(f"File not found: {api_key_path}")

        self.portfolio_cache = PortfolioCache()
        self.portfolios = self._get_portfolios()

        self.market_returns = self._get_market().tz_localize(None)

        self.rf_rate = self._get_rf_rate()

    def _get_portfolios(self) -> list:
        """
        Get Portfolio objects from string combinations.
        """
        portfolios = []
        for comb in self.ncr_gen:
            portfolio = Portfolio(*comb)
            portfolios.append(portfolio)
        return portfolios

    def _get_market(self) -> pd.Series:
        """Get the market returns for the given horizon."""
        start = dt.datetime.today() - dt.timedelta(days=self.lookback)
        start = start.date()
        end = dt.datetime.today().date()

        market_close = yf.Ticker(self.market).history(start=start, end=end)["Close"]
        market_returns = (market_close - market_close.shift(self.horizon)) / market_close.shift(self.horizon)
        market_returns = market_returns.dropna()

        return market_returns

    def _get_rf_rate(self) -> float:
        """Get the risk-free rate for the given horizon."""
        ticker = yf.Ticker("^TNX")
        pa_rate = ticker.history(period="1d")["Close"].iloc[0] / 100

        horizon_rate = (1 + pa_rate / 2)**(2*(self.horizon / 365)) - 1  # Semi-annual compounding with ACT/365 convention
        return horizon_rate

    def _iteration_optimize(self, portfolio: Portfolio, bounds: tuple[float, float] = (0.0, 1.0)) -> dict[str, Any]:
        """Optimization function ran in parallel iteration of portfolios.

        :param portfolio: Portfolio object
        :return: tuple of Portfolio object and optimized weights
        """
        # Cache query
        cached = self.portfolio_cache.get(portfolio, self.target, bounds)
        if cached:
            return cached

        # Get the historical data for the portfolio
        periodic_returns = self.periodic_returns.loc[:, list(portfolio)]
        historical_close = self.historical_close.loc[:, list(portfolio)]

        return_preds = ReturnPred(historical_close, self.horizon).all_stocks_pred(comb=True)
        expected_returns = np.array([return_dict['expected_return'] for return_dict in return_preds.values()])

        if self.key_path and self.key_var and self.sentiment_analysis:
            sentiment = Sentiment(self.key_path, self.key_var)
            sentiment_period = min(30, self.horizon)

            sentiment_adjustment = [sentiment.get_sentiment(stock, 3, sentiment_period) for stock in portfolio]
            expected_returns = expected_returns * (1 + 0.33 * np.array(sentiment_adjustment))

        cov_matrix = periodic_returns.cov()

        betas = np.array([
            np.cov(
                periodic_returns[stock].align(self.market_returns, join='inner')[0],  # Align the two series
                self.market_returns.align(periodic_returns[stock], join='inner')[0]
            )[0, 1] / np.var(self.market_returns.align(periodic_returns[stock], join='inner')[0])
            for stock in portfolio
        ])

        # Constraints
        def check_sum(weights):
            return np.sum(weights) - 1

        def target_return(weights):
            return np.dot(weights, expected_returns) - self.target

        initial_guess = np.array([1/len(portfolio) for _ in range(len(portfolio))])
        bounds_ls = [bounds for _ in range(len(portfolio))]
        constraints = [{'type': 'eq', 'fun': check_sum},
                       {'type': 'ineq', 'fun': target_return}]  # 'ineq' will produce return >= target as a constraint

        # Objective function
        def objective(weights):
            return weights.T @ cov_matrix @ weights

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            result = minimize(objective, initial_guess, bounds=bounds_ls, constraints=constraints)  # type: ignore

        out = {
            "portfolio": " - ".join(portfolio),
            "weights": result.x.round(4).tolist(),  # Convert to list
            "return": np.dot(result.x, expected_returns).round(4),
            "portfolio_variance": result.fun.round(4),
            "expected_returns": expected_returns.tolist(),  # Convert to list
            "cov_matrix": cov_matrix.values.tolist(),  # Convert to a nested list
            "betas": betas
        }

        self.portfolio_cache.cache(portfolio, self.target, bounds, out)
        return out

    def optimize_space(self, bounds: tuple = (0.0, 1.0)) -> nCrResult:
        """
        Optimize the combination space.

        :return: List of optimized portfolios (best to worst)
        """
        results = nCrResult([self._iteration_optimize(portfolio, bounds)
                            for portfolio in tqdm(self.portfolios, desc="Iterating over portfolios")],
                            rf_rate=self.rf_rate)

        results.sort(key=lambda x: (x['return'], -x['portfolio_variance']), reverse=True)

        return results



from typing import Union, Iterable
import yfinance as yf
import math

from .CustomTypes import StockSymbol

class Portfolio(tuple):
    def __new__(cls, *args):
        obj = super().__new__(cls, sorted(args))

        obj._weights = [1/len(obj) for _ in range(len(obj))]

        obj._tickers = yf.Tickers(' '.join(obj))

        obj._results = {
            'weights': {obj[i]: -1 for i in range(len(obj))},
            'expected_returns': {obj[i]: -1 for i in range(len(obj))},
            'volatility': {obj[i]: -1 for i in range(len(obj))},
            'beta': {obj[i]: -1 for i in range(len(obj))},
            'sharpe_ratio': {obj[i]: -1 for i in range(len(obj))},
            'sentiment': {obj[i]: -1 for i in range(len(obj))}
        }

        obj.optimum_portfolio_info = {
            'target_return': None,
            'target_volatility': None,
            'weights': None,
        }

        return obj

    def _stock_results(self, stock: StockSymbol):
        return {
            'weight': self._results['weights'][stock],
            'expected_return': self._results['expected_returns'][stock],
            'volatility': self._results['volatility'][stock],
            'beta': self._results['beta'][stock],
            'sharpe_ratio': self._results['sharpe_ratio'][stock]
        }

    def __getitem__(self, key: Union[int, StockSymbol]):
        if isinstance(key, int):
            return super().__getitem__(key)

        elif isinstance(key, StockSymbol):
            return self._stock_results(key)

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, sub_category: str, value: dict):
        assert sub_category in self._results.keys(), f"Invalid sub_category: {sub_category}, must be one of {self._results.keys()}"
        self._results[sub_category] = value

    @property
    def tickers(self):
        return self._tickers

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, value: Iterable[float]):
        assert len(value) == len(self), f"Invalid number of weights: {len(value)}, must be {len(self)}"
        self._weights = value




from datetime import timedelta, datetime as dt
import os
import warnings
from dotenv import load_dotenv

from .Cache import SentimentCache

from newsapi import NewsApiClient

import pandas as pd

import torch
from transformers import BertTokenizer, BertForSequenceClassification

from .CustomTypes import Days


class Sentiment:
    """
    FinBERT Sentiment Analysis for Financial News.
    """
    def __init__(self, api_key_path: os.PathLike, api_key_var: str) -> None:
        warnings.filterwarnings("ignore")
        if not os.path.exists(api_key_path):
            raise FileNotFoundError(api_key_path)

        self.set_api_key(api_key_path)

        if os.getenv(api_key_var) is None:
            raise ValueError(f"{api_key_var} not found in environment variables")

        self.key_var = api_key_var

        self.tokenizer = BertTokenizer.from_pretrained('yiyanghkust/finbert-tone')
        self.model = BertForSequenceClassification.from_pretrained('yiyanghkust/finbert-tone')
        self.cache = self._init_cache()

    @staticmethod
    def set_api_key(path: os.PathLike | str) -> None:
        if not str(path).split('.')[-1] == 'env':
            raise ValueError("Invalid file type. Must be a .env file")

        load_dotenv(path)

    @staticmethod
    def _init_cache(name: str = 'sentiment.db', exp_seconds: int = 3600) -> SentimentCache:
        return SentimentCache(name, exp_seconds)

    def search(self, query: str, *, n: int, lookback: Days) -> list:
        key = os.getenv(self.key_var)
        newsapi = NewsApiClient(api_key=key)

        date = dt.today() - timedelta(days=lookback)

        articles = newsapi.get_everything(q=query,
                                          from_param=date,
                                          language='en',
                                          sort_by='publishedAt',
                                          page_size=n)

        filtered_articles = []

        neutral_string = 'confident.' # Evaluates to a sentiment polarity of 0.0 (neutral). 200 IQ solution.

        for article in articles['articles']:
            if (article['description'] is None or article['description'] == '') and \
                    (article['title'] is None or article['title'] == ''):
                continue  # Skip articles with both description and title as None or empty

            # Ensure no None values remain
            if article['description'] is None:
                article['description'] = neutral_string
            if article['title'] is None:
                article['title'] = neutral_string

            filtered_articles.append(article)

        desc = [article['description'] + ' ' + article['title'] for article in filtered_articles]

        if not desc:
            raise ValueError("No articles found for the given query")

        return desc

    def get_score_all(self, text: str) -> dict:

        assert text and isinstance(text, str), "Input must be a non-empty string"

        inputs = self.tokenizer(text,
                                return_tensors='pt',
                                padding='max_length',
                                truncation=True,
                                max_length=512)

        outputs = self.model(**inputs)

        logits = outputs.logits.squeeze()
        p_val = torch.nn.functional.softmax(logits, dim=0).detach().numpy()
        return p_val

    def compose_sentiment(self, text: str) -> str:
        p_val = self.get_score_all(text)

        score = p_val[1] - p_val[2]

        return score

    def get_sentiment(self, query: str, n: int, lookback: Days) -> float:
        cache_query = f"{query} {lookback=} {n=}"

        cache_response = self.cache.get(cache_query)

        # Cache hit
        if cache_response is not None:
            return cache_response

        # Cache miss
        search_results = self.search(query, n=n, lookback=lookback)
        sentiments = []

        for desc in search_results:
            score = self.compose_sentiment(desc)
            sentiments.append(score)

        if not sentiments:
            self.cache.cache(cache_query, 0.0)
            return .5  # Neutral if no sentiment found

        ewma_sentiment = pd.Series(sentiments).ewm(halflife=2).mean().iloc[-1]
        self.cache.cache(cache_query, ewma_sentiment)
        return ewma_sentiment





# ML imports
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error, mean_absolute_percentage_error
from sklearn.model_selection import GridSearchCV

# Data imports
import pandas as pd
import numpy as np

# Typing imports
from typing import Union
from .CustomTypes import Days


class ReturnPred:
    """Create expected return predictions from historical prices"""

    def __init__(self, data: pd.DataFrame, inv_horizon: Days = 21):
        self.model = RandomForestRegressor()
        self.data = data
        self.inv_horizon = inv_horizon

    def split_stocks(self):
        """Split a DataFrame of multiple stocks into individual DataFrame"""
        out = []
        for col in self.data.columns:
            out.append(self.data[col])
        return out

    def add_lagged_features(self, stock_data: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        """
        Create lagged features for a stock's historical data and set the target as the future price.

        Parameters:
            stock_data (Union[pd.DataFrame, pd.Series]): A time-series of stock prices or returns.

        Returns:
            pd.DataFrame: A DataFrame with lagged features and the original target column (future price).
        """
        if isinstance(stock_data, pd.Series):
            stock_data = stock_data.to_frame(name="future_price")

        assert stock_data.shape[1] == 1, "stock_data must have exactly one column"

        # Calculate lag component based on inv_horizon
        lag_component = max(4, int(np.floor(np.sqrt(self.inv_horizon))))

        feature_set = pd.DataFrame(index=stock_data.index)

        # Create lagged features
        for i in range(1, lag_component + 1):  # Include lag_component in the loop
            feature_set[f"lag_{i}"] = stock_data.shift(i).values.ravel()

        # Calculate the future price at i+inv_horizon
        future_price = stock_data.shift(-self.inv_horizon)

        # Combine features and target (future price), dropping rows with NaNs due to lagging
        data = pd.concat([feature_set, future_price], axis=1)

        today_data = data.iloc[-1] # Last days data for return prediction (future_price is NaN)
        today_data = today_data.to_frame().dropna().T

        data = data.dropna(how="any")
        return data, today_data


    def train(self, stock_data: Union[pd.DataFrame, pd.Series], comb: bool = False) -> dict:
        """Train the model on a stock's historical returns. Hyperparameters are not tuned in order to keep the runtime
        low. In case of unacceptable accuracy, the fallback method is an historical EWMA with a span equal to the
        investment horizon (in days). This is an iteration on the traditional return calculation of using the mean return
        over the data period.

        :param stock_data: A time-series of stock prices or returns.
        :param comb: Boolean to indicate if the portfolio is part of a combination space from `nCrEngine`.
        """

        success = False
        confidence = 0
        data, today_data = self.add_lagged_features(stock_data)

        X = data.drop(columns=data.columns[-1])
        y = data[data.columns[-1]]

        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=0)

        # Simple Grid-Search
        if not comb:
            param_grid = {
                'n_estimators': [100, 200],
                'max_depth': [None, 10, 20],
                'min_samples_split': [2, 5],
                'max_features': ['sqrt', 'log2'],
                'bootstrap': [True]
            }
            gscv = GridSearchCV(self.model, param_grid, cv=5, n_jobs=-1, scoring='neg_mean_squared_error')
            gscv.fit(X_train, y_train)

            self.model = gscv.best_estimator_
        elif comb:
            self.model.fit(X_train, y_train)

        pred = self.model.predict(X_test)
        target_error = np.var(y_test)

        if mean_squared_error(y_test, pred) < target_error:
            success = True

        confidence = np.round(1 - mean_absolute_percentage_error(y_test, pred), 4)

        # Return prediction for the period r_{t:t+inv_horizon}
        expected_price = np.round(self.model.predict(today_data)[0], 2)
        expected_return = np.round((expected_price - stock_data.iloc[-1]) / stock_data.iloc[-1], 4)

        if not success:
            period_returns = (stock_data - stock_data.shift(self.inv_horizon)) / stock_data.shift(self.inv_horizon)

            span = np.round(np.sqrt(self.inv_horizon), 0) * 2
            expected_return = np.round(period_returns.ewm(span=span).mean().iloc[-1], 4)
            expected_price = np.round(stock_data.iloc[-1] * (1 + expected_return), 2)

        return {'success': success,
                'confidence': confidence,
                'expected_price': expected_price,
                'expected_return': expected_return}


    def all_stocks_pred(self, comb: bool = False) -> dict:
        """Return predictions for all stocks in the data"""
        stocks = self.split_stocks()
        return {stock.name: self.train(stock, comb) for stock in stocks}


from .CacheConstructor import CacheConstructor

from pandas import DataFrame
from datetime import datetime as dt, timedelta
import pickle

from typing import Union, Optional, Any


class SentimentCache(CacheConstructor):

    def __init__(self, name: Optional[str] = None, exp_seconds: int = 3600) -> None:
        super().__init__(name or "sentiment")
        self.exp_seconds: int = exp_seconds

        self.create({
            "symbol": "TEXT PRIMARY KEY",
            "sentiment": "REAL",
            "createdAt": "TEXT",
            "expireAfter": "INTEGER"
        })

    def cache(self, symbol: str, sentiment: float) -> None:
        now = dt.now().isoformat()
        expire = self.exp_seconds
        q = f"INSERT OR REPLACE INTO {self.name} (symbol, sentiment, createdAt, expireAfter) VALUES (?, ?, ?, ?)"
        self.exec(q, (symbol, sentiment, now, expire))

    def get(self, symbol: str) -> Optional[Union[dict, float]]:
        # Check if the symbol is in the cache
        q = f"SELECT * FROM {self.name} WHERE symbol = ?"
        self.exec(q, (symbol,))
        data = self.curr.fetchone()

        if data is None:
            return None

        # Check if the data is expired
        now = dt.now()
        expire = dt.fromisoformat(data[2]) + timedelta(seconds=data[3])
        if now - expire > timedelta(seconds=0):
            return None

        return data[1]


class PortfolioCache(CacheConstructor):
    def __init__(self, name: Optional[str] = None, expire_days: int = 1) -> None:
        super().__init__(name or "portfolio")
        self.expire_days = expire_days

        self.create({
                "portfolio": "TEXT PRIMARY KEY",
                "data": "BLOB",
                "createdAt": "DATE",
                "expiresAt": "DATE"
            })

    def cache(self, portfolio: tuple, target_return: float, bounds: tuple[float, float], data: Any) -> None:
        portfolio_id = f'({", ".join(portfolio)})_{target_return}_{bounds}'
        created_at = dt.today().date()

        expires_at = created_at + timedelta(days=self.expire_days)

        data = pickle.dumps(data)
        self.exec(
            f"INSERT OR REPLACE INTO {self.name} (portfolio, data, createdAt, expiresAt) VALUES (?, ?, ?, ?)",
            (portfolio_id, data, created_at, expires_at))


    def get(self, portfolio: tuple, target_return: float, bounds: tuple[float, float]) -> Any:
        portfolio_id = f'({", ".join(portfolio)})_{target_return}_{bounds}'

        self.exec(f"""SELECT data, expiresAt FROM {self.name}
                             WHERE portfolio=?""", (portfolio_id,))
        response = self.curr.fetchone()

        if response:
            data, expires_at = response
            if dt.now() < dt.strptime(expires_at, "%Y-%m-%d"):
                return pickle.loads(data)
            else:
                # Cache expired, remove the entry
                self.exec(f"DELETE FROM {self.name} WHERE portfolio=?",
                          (portfolio_id,))

        return None


class nCrCache(CacheConstructor):
    def __init__(self, name: Optional[str] = None, expire_days: int = 1) -> None:
        super().__init__(name or "nCr")

        self.expire_days = expire_days

        self.create({
            "id": "TEXT PRIMARY KEY",
            "data": "BLOB",
            "createdAt": "DATE",
            "expiresAt": "DATE"
        })

    def cache(self, _id: str, data: Any) -> None:
        """
        Cache the data.

        :param _id: identifier for with format: f"{market_name}_{lookback}"
        :param data: BLOB of price data
        """

        created_at = dt.now().date()
        expires_at = created_at + timedelta(days=self.expire_days)

        data = pickle.dumps(data)

        query = f"""
        INSERT OR REPLACE INTO {self.name} (id, data, createdAt, expiresAt)
        VALUES (?, ?, ?, ?)
        """
        self.exec(
            query,
            (_id, data, created_at, expires_at))

    def get(self, _id: str) -> Optional[DataFrame]:
        """
        Get the data from the cache.

        :param _id: identifier for with format: f"{market_name}_{lookback}"
        """
        query = f"""
        SELECT * FROM {self.name} WHERE id = ?
        """
        self.exec(query, (_id,))
        data = self.curr.fetchone()

        if data is None:
            return None

        if dt.fromisoformat(data[3]).date() < dt.now().date():
            return None
        else:
            return pickle.loads(data[1])



from typing import NewType, Literal, Tuple, Union, NewType

# Type aliases
Days = int

StockSymbol = str
StockDataSubset = Tuple[Literal['Open', 'High', 'Low', 'Close', 'Volume', 'Dividends', 'Stock Splits']]
IndexSymbol = str


import sqlite3 as sql
from typing import Optional, Any


class CacheConstructor:

    def __init__(self, name: str) -> None:
        self.name = name
        self.conn, self.curr = self._connect()

    @staticmethod
    def _connect() -> tuple[sql.Connection, sql.Cursor]:
        conn = sql.connect('cache.db')
        curr = conn.cursor()
        return conn, curr

    def pass_connection(self) -> tuple[sql.Connection, sql.Cursor]:
        return self.conn, self.curr

    def exec(self, query: str, params: Optional[tuple] = None) -> Optional[Any]:
        try:
            # Execute the query with parameters
            if params:
                self.curr.execute(query, params)
            else:
                self.curr.execute(query)

            # Commit changes for write operations
            self.conn.commit()

            # Return results for read operations
            if query.strip().lower().startswith("select"):
                return self.curr.fetchall()

            return None

        except Exception as e:
            print(f"Error executing query: {query}\nParameters: {params}\nError: {e}")
            return None

    def create(self, cols: dict[str, str]) -> None:
        query = f"CREATE TABLE IF NOT EXISTS {self.name} ("
        for col, dtype in cols.items():
            query += f"{col} {dtype}, "
        query = query[:-2] + ")"

        self.exec(query)

    def clear(self) -> None:
        self.exec(f"DELETE FROM {self.name}")


    def close(self) -> None:
        self.conn.close()

    def __del__(self) -> None:
        self.close()



from .Cache import nCrCache
from .CustomTypes import IndexSymbol, Days, StockSymbol

import pandas as pd
import numpy as np

from itertools import combinations

import yfinance as yf
from json import loads
import requests
from bs4 import BeautifulSoup
from io import StringIO
import datetime as dt

from typing import Optional

from scipy.optimize import minimize


class nCrEngine:
    """
    Class to perform portfolio selection from all possible nCr combinations of a list of index components.
    """

    def __init__(self, market: IndexSymbol,
                 n: int = 5,
                 horizon: Days = 21,
                 lookback: Days = 252,
                 max_pool_size: Optional[int] = None,
                 target_return: float = 0.1):
        assert n > 0, "n must be greater than 0"
        assert lookback > horizon > 0, "Lookback must be greater than horizon and both must be greater than 0"

        with open("NeoPortfolio/INDEX_MAP.json", "r") as f:
            self.INDEX_MAP = loads(f.read())

        if market not in self.INDEX_MAP.keys():
            raise ValueError(f"Invalid market: {market}, must be one of {self.INDEX_MAP.keys()}")

        self.cache = nCrCache(expire_days=1)

        self.horizon = horizon
        self.lookback = lookback

        self.target = target_return

        self.market = market
        self.n = n

        self.components = self._get_components(market)
        self.tickers = self._get_tickers(self.components)

        self.historical_close = self._get_historical_close(lookback=lookback)
        self.periodic_returns = self._get_periodic_returns(horizon=horizon)
        self.expected_returns = self._get_ewma_expected(horizon=horizon)
        self.volatility = self._get_volatility(horizon=horizon)

        self.comb_space_count = min(self._get_comb_space_count(), max_pool_size)
        self.high_ret = self._high_return_stock_proportion(n)
        self.low_vol = 1 - self.high_ret

        self.combination_space = self._compose_combination_space()
        self.ncr_gen = self._get_nCr_generator(self.combination_space, n)

    def _get_components(self, market: IndexSymbol) -> list[StockSymbol]:
        """
        Get the components of the index.

        :param market: str | Symbol of the index to take as the market.
        """
        url = self.INDEX_MAP[market]['url']
        table_index = self.INDEX_MAP[market]['table_index']
        col = self.INDEX_MAP[market]['col']

        response = requests.get(url)
        response.raise_for_status()

        soup = BeautifulSoup(response.text, 'html.parser')
        table = soup.find_all('table')[table_index]
        table = StringIO(str(table))
        df = pd.read_html(table)[0]

        components = pd.Series(df[col])

        # Fix known exceptions
        if components.isin(["BRK.B"]).any():
            components = components.replace("BRK.B", "BRK-B")

        if components.isin(["BF.B"]).any():
            components = components.replace("BF.B", "BF-B")

        return components.tolist()

    @staticmethod
    def _get_nCr_generator(components: list[StockSymbol], n: int) -> list[tuple[StockSymbol]]:
        """
        Get all possible nCr combinations of a list of components.

        :param components: list | List of components to take combinations from.
        :param n: int | Number of components to take in each combination.
        :return: list | Generator of all possible nCr combinations of the components.
        """
        for comb in combinations(components, n):
            yield comb

    def _get_tickers(self, components: list[StockSymbol]) -> yf.Tickers:
        """
        Get yf.Tickers object from a list of components.

        :param components: list | List of components to get the Tickers object for.
        :return: yf.Tickers | Tickers object for the components.
        """
        assert self.components, "Components could not be found."

        return yf.Tickers(' '.join(components))

    def _get_historical_close(self, lookback: Days) -> pd.DataFrame:
        """
        Get historical close prices for all components.

        :param lookback: int | Number of days to look back for historical data.
        :return: pd.DataFrame | Historical close prices for all components.
        """
        assert self.components, "Components could not be found."

        # Cache Check
        query_id = f"{self.market}_{lookback}"
        response = self.cache.get(query_id)

        if response is not None:
            return response

        start = dt.datetime.today() - dt.timedelta(days=lookback)
        start = start.date()

        end = dt.datetime.today().date()

        data = yf.download(' '.join(self.components), start=start, end=end)['Close']
        self.cache.cache(query_id, data)

        return data

    def _get_periodic_returns(self, horizon: Days) -> pd.DataFrame:
        """
        Get periodic returns for all components.

        :param horizon: int | Number of days to look forward for periodic returns.
        :return: pd.DataFrame | Periodic returns for all components.
        """
        periodic_returns = self.historical_close.copy()

        periodic_returns = (periodic_returns - periodic_returns.shift(horizon)) / periodic_returns.shift(horizon)
        periodic_returns = periodic_returns.dropna()

        return periodic_returns

    def _get_ewma_expected(self, horizon: Days) -> pd.Series:
        """
        Get the exponentially weighted moving average expected returns.

        :param horizon: int | Number of days to look forward for expected returns.
        :return: pd.Series | Expected returns for all components.
        """
        ewma_expected = self.periodic_returns.ewm(halflife=horizon).mean().iloc[-1]
        return ewma_expected

    def _get_volatility(self, horizon: Days) -> pd.Series:
        """
        Get the volatility of the components.

        :param horizon: int | Number of days to look forward for volatility.
        :return: pd.Series | Volatility of the components.
        """
        ewma_var = self.periodic_returns.pow(2).ewm(span=horizon).mean()
        volatility = ewma_var.apply(np.sqrt).iloc[-1]
        return volatility

    @staticmethod
    def _high_return_stock_proportion(n: int = 5) -> float:
        """
        Get the proportion of high return stocks in the portfolio.

        :param n: int | Number of components in the portfolio.
        :return: float | Proportion of high return stocks in the portfolio.
        """
        return 0.7/(1+(n-5)/5)  # Assume 5 to be the average portfolio size and 0.7
                                # to be the proportion of high return stocks in the average portfolio

    def _get_comb_space_count(self) -> int:
        """
        Get the number of combinations in the combination space.

        :return: int | Number of combinations in the combination space.
        """
        comb_space_count = min(
        max(self.n * round(np.log(len(self.components))), 2 * round(np.log(len(self.components)))),
        len(self.components)
        )
        return comb_space_count

    def _compose_combination_space(self) -> list[StockSymbol]:
        """
        Compose the combination space of all possible nCr combinations of components.

        :return: list | All members of the combination components.
        """
        hi_ret = round(self.high_ret * self.comb_space_count)
        lo_vol = round(self.low_vol * self.comb_space_count)

        df = pd.DataFrame({
            'expected_return': self.expected_returns,
            'volatility': self.volatility
        })
        df = df.loc[df['expected_return'] > 0, :]

        ret_based = df.loc[df['expected_return'] <= 2 * self.target, 'expected_return'].nlargest(hi_ret).index
        vol_based = df.loc[df['volatility'] < df['expected_return'], 'volatility'].nsmallest(lo_vol).index
        return list({*ret_based, *vol_based})

    def pass_optimization_params(self) -> tuple[int, Days, Days, float, pd.DataFrame, pd.DataFrame, pd.Series]:
        """
        Pass the optimization parameters to the nCrOptimize object.

        :return: tuple | Parameters to pass to the nCrOptimize object.
        """
        return (self.n,
                self.horizon,
                self.lookback,
                self.target,
                self.historical_close,
                self.periodic_returns,
                self.volatility)

